use std::{collections::VecDeque, env, path::Path};

use crate::{token::{Token, MessageType}, builtin::modules::Module};
use crate::builtin::modules::io::print;

#[derive(Default)]
pub struct LangSpecs {
    /// builtin modules for extra builtin functions
    builtin_features: Vec<crate::builtin::modules::Module>,
    lang_version: u32,
    authors: Vec<String>,
    embedded_files: Vec<(usize, String, String)>,
}

impl LangSpecs {
    
    pub fn features(&self) -> &[crate::builtin::modules::Module] {
        &self.builtin_features
    }

    pub fn embedded_files(&self) -> &[(usize, String, String)] {
        &self.embedded_files
    }
}

pub fn resolve_directives(tokens: &mut VecDeque<Token>, origin: &String) -> LangSpecs {
    let mut specs = LangSpecs::default();

    for (idx, token) in tokens.iter().enumerate() {
        match token {
            Token::CompilerDirective(text, _) => parse_directive(text, idx, &mut specs, origin),
            _ => ()
        }
    }

    specs
}

static DIRECTIVE_REGEX_SRC: &'static str = concat!(
    r"@feature\(((?:\s*[\w]+\s*,?)*)\)",
    r"|@version\(\s*([0-9]{3})\s*\)",
    r"|@author\((.*)\)",
    r"|@include\((.*)\)"
);

lazy_static::lazy_static! {
    static ref DIRECTIVE_REGEX: regex::Regex = regex::Regex::new(DIRECTIVE_REGEX_SRC).unwrap();
}

pub fn from_list(text: &str) -> Vec<String> {
    let mut vec = vec![];

    for word in text.split(',') {
        vec.push(String::from(word.trim()))
    }

    vec
}

fn parse_directive(text: &str, token_idx: usize, specs: &mut LangSpecs, origin: &String) {

    for cap in DIRECTIVE_REGEX.captures_iter(text) {
        let mut enumerator = cap.iter().enumerate();
        loop {
            let next = enumerator.next();
            if next.is_none() {
                break;
            }

            let (i, group) = next.unwrap();

            // ignore first group as its the entire match,
            // as well as the 1st group (= comments)
            if i < 1 {
                continue;
            }

            if let Some(mat) = group {
                match i {
                    1 => {
                        specs.builtin_features.append(&mut Module::from_list(mat.as_str()));
                        return;
                    },
                    2 => {
                        specs.lang_version = mat.as_str().parse().unwrap();
                        return;
                    },
                    3 => {
                        specs.authors.push(String::from(mat.as_str()));
                        return;
                    },
                    4 => {
                        let cwd = Path::new(origin).parent().unwrap().canonicalize().unwrap();
                        for path in from_list(mat.as_str()).iter() {
                            let mut include_path = cwd.clone();
                            include_path.push(path);

                            if let Ok(path) = include_path.canonicalize() {
                                let include_file = path.to_str().unwrap().to_string();
                                
                                if let Ok(str) = std::fs::read_to_string(&include_file) {
                                    specs.embedded_files.push((token_idx, str, include_file));
                                } else {
                                    crate::message(MessageType::Warning, format!("Unable to read embedded file: {include_file}"));
                                }
                            }
                        }
                        return;
                    },
                    _ => crate::message(crate::token::MessageType::Warning, format!("unknown directive: `{}`", text)),
                }
                break;
            }
        }
    }
    
    crate::message(crate::token::MessageType::Warning, format!("unknown directive: `{}`", text));
}